#!/usr/bin/env python
#
# file: /data/isip/exp/demos/exp_0007/v0.0.6/imld_alg_lda.py
#
# revision history:
# 20200811 (LV): standardization, completion
# 20200505 (SJ): initial version
#
# This script implements the Class Dependent and Independent Linear
# Discriminant Analysis machine learning algorithm for the ISIP Machine
# Learning Demo software.
#-----------------------------------------------------------------------------

# import system modules
#
import numpy as np

# import data related modules
#
from scipy.spatial import distance
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis

# set error handling for floating point errors
#
np.seterr(divide='ignore', invalid='ignore')

# class: AlgorithmLDA_ci
#
# This class contains methods for the Class-Independent Linear D
#
class AlgorithmLDACI():
    def __init__(self,
                 input_display,
                 eval_display,
                 process_box):

        self.input = input_display
        self.eval = eval_display
        self.process = process_box
        self.class_num = np.empty((0, 0))
        self.X = np.empty((0, 0))
        self.classes = 0
        self.mu = None
        self.lda = LinearDiscriminantAnalysis(solver="eigen")
        self.ldas = [LinearDiscriminantAnalysis(solver='eigen') for i in \
                     range(4)]
        self.support_region = np.empty((0, 0))
        self.cov1 = 0
        self.cov2 = 0
        self.LDA = np.empty((2, 2)) * np.nan

        self.trans_matrix = np.empty((0, 0))
        self.step_index = 0

        self.data = False

    def run(self):
        if self.step_index == 1:
            self.step1()

        elif self.step_index == 2:
            self.step2()
            self.process.append("\n   *** Algorithm Complete ***\n")

    def get_step_index(self):
        if self.step_index > 2:
            finished = True
        else:
            finished = False

        return finished

    def increment_step(self):
        self.step_index = self.step_index + 1

    def reset_step(self):
        self.step_index = 0

    def initialize(self):
        text = "Algorithm: Class Independent Principal Component Analysis"
        self.process.append(text)
        text = "\n    *** Initializing the Algorithm ***\n"
        self.process.append(text)

        self.class_num = np.empty((0, 0))
        self.X = np.empty((0, 0))
        self.classes = 0

        self.lda = LinearDiscriminantAnalysis(solver="eigen")
        self.support_region = np.empty((0, 0))

        self.LDA = np.empty((2, 2)) * np.nan
        self.trans_matrix = np.empty((0, 0))

        self.step_index = 0
        self.data = False

    def step1(self):
        text = "    1. Computing the Statistics.\n" + "       Means:"
        self.process.append(text)

        self.print_means()
        self.input.canvas.draw_idle()
        self.print_matrices()

    def step2(self):
        text = "\n    2. Drawing the Decision Surface(s).\n "
        self.process.append(text)

        self.euclidean_classifier(self.X, self.input, self.classes, self.lda)

        text = "    3. Computing Error Rates:\n "
        self.process.append(text)

        text = "    Train:"
        self.process.append(text)
        self.compute_errors(self.input)

        text = "    Eval:"
        self.process.append(text)
        self.classify_eval()

        self.input.reset_classes()

    def extract_data(self):
        self.classes = 0
        index = 0
        observed_classes = np.empty((0, 0))

        if len(self.input.canvas.axes.collections) == 0:
            return False
        else:
            self.data = True

        index_col = np.empty((0, 0), dtype="int64")
        collection_index = 0
        for collections in self.input.canvas.axes.collections:
            # retrieve the x,y and gid of the collection
            #
            try:
                current_gid = collections.get_gid()[0][0]
            except TypeError:
                current_gid = collections.get_gid()

            if current_gid is not None:
                data = np.array(collections.get_offsets())
                classes = np.array(collections.get_gid())

                self.class_num = np.append(self.class_num, classes)
                self.X = np.append(self.X, data)

                if current_gid not in observed_classes:
                    observed_classes = np.append(observed_classes, current_gid)
                    label = np.reshape(np.array(collections.get_gid()), (-1,))
                    self.classes = self.classes + 1
                    index = index + 1
                    index_col = np.append(index_col, collection_index)

                else:
                    pca_index = 0
                    for collects in self.input.canvas.axes.collections:
                        if current_gid in collects.get_gid():
                            save_sets = collects.get_offsets()
                            current_set = collections.get_offsets()
                            final_set = np.vstack((save_sets, current_set))

                            save_gid = collects.get_gid()
                            cur_gid = collections.get_gid()
                            final_gid = np.hstack((save_gid, cur_gid))

                            collects.set_offsets(final_set)
                            collects.set_gid(final_gid)
                            break
                        else:
                            pca_index = pca_index + 1
            collection_index = collection_index + 1

        remove_collections = np.array(self.input.canvas.axes.collections)[index_col].tolist()
        self.input.canvas.axes.collections = remove_collections

        self.X = np.reshape(self.X, (-1, 2))

        self.class_num = np.ndarray.flatten(self.class_num)

        self.transformLDA(self.X, self.class_num, self.classes)

        return True

    def transformLDA(self, X, y, classes):
        self.lda.fit_transform(X, y)

        if X.size > 0:
            mu = np.empty((0, 0))
            for num in range(classes):
                index = np.where(y == num)[0]
                reshaped_x = np.reshape(X[index], (-1, 2))
                mu_v = np.mean(reshaped_x, axis=0)
                mu = np.append(mu, mu_v)

            mu = np.reshape(mu, (classes, 2, 1))
            self.mu = mu

        eigVect = self.lda.scalings_

        for j in range(2):
            for k in range(2):
                self.LDA[k][j] = eigVect[k][j]

    def euclidean_classifier(self, X, display, classes, lda):

        res = (display.canvas.axes.get_xlim()[1] - display.canvas.axes.get_ylim()[0]) / 100

        x_min, x_max = X[:, 0].min() - .75, X[:, 0].max() + .75
        y_min, y_max = X[:, 1].min() - .75, X[:, 1].max() + .75
        xx, yy = np.meshgrid(np.arange(x_min, x_max, res),
                             np.arange(y_min, y_max, res))

        Z = np.empty((0, 0))

        for pairs in np.c_[xx.ravel(), yy.ravel()]:
            distances = np.empty((0, 0))
            for data in range(classes):

                mean = np.reshape(self.mu[data], (-1,1))

                trans_mean = np.matmul(self.LDA,mean)
                trans_pair = np.matmul(self.LDA,pairs)

                eud = (distance.euclidean(trans_mean, trans_pair))
                distances = np.append(distances, eud)

            Z = np.append(Z, np.argmin(distances))

        Z = Z.reshape(xx.shape)

        display.canvas.axes.contourf(xx, yy, Z, alpha=0.4)
        display.canvas.draw_idle()

    def print_means(self):
        for means in range(self.classes):
            text = "       Class {}:{}".format(means + 1, np.reshape(self.mu[means], (2,)))
            self.input.canvas.axes.scatter(self.mu[:, 0], self.mu[:, 1], c='black', s=5)

            self.process.append(text)

    def print_matrices(self):

        cov = self.lda.covariance_

        a11 = str(round(cov[0][0], 2))
        a12 = str(round(cov[0][1], 2))
        a21 = str(round(cov[1][0], 2))
        a22 = str(round(cov[1][1], 2))

        text = ("\n      Covariance matrix:\n" +
                "         " + a11 + "    " + a12 + "\n" +
                "        " + a21 + "     " + a22 + "\n")

        self.process.append(text)
        a11 = str(round(self.LDA[0][0], 2))
        a12 = str(round(self.LDA[0][1], 2))
        a21 = str(round(self.LDA[1][0], 2))
        a22 = str(round(self.LDA[1][1], 2))

        text = ("      Transformation matrix:\n" +
                "         " + a11 + "    " + a12 + "\n" +
                "         " + a21 + "    " + a22 + "\n")
        self.process.append(text)


    def euclidean_classifier_point(self, data, lda):

        distances = np.empty((0, 0))
        for cur_class in range(self.classes):
            mean = np.reshape(self.mu[cur_class], (-1, 1))

            trans_mean = np.matmul(self.LDA, mean)
            trans_pair = np.matmul(self.LDA, data)

            eud = (distance.euclidean(trans_mean, trans_pair))
            distances = np.append(distances, eud)

        return np.argmin(distances)

    def compute_errors(self, display):
        # initialize variables
        #
        samples = 0

        incorrect1 = 0
        incorrect2 = 0
        incorrect3 = 0
        incorrect4 = 0

        for collection in range(self.classes):
            size = display.canvas.axes.collections[collection].get_offsets().shape[0]
            samples = samples + size

        text = "    Number of samples: {}".format(samples)
        self.process.append(text)

        if self.classes == 4:
            text = "    Ref/Hyp:    {:10s}{:10s}{:10s}{:10s}".format("Class 1", "Class 2", "Class 3", "Class 4")
        elif self.classes == 3:
            text = "    Ref/Hyp:    Class 1       Class 2        Class 3"

        elif self.classes == 2:
            text = "    Ref/Hyp:    {:10s}{:10s}".format("Class 1", "Class 2")

        elif self.classes == 1:
            text = "    Ref/Hyp:    {:10s}".format("Class 1")

        self.process.append(text)

        # computes the classification error for the first set
        #
        if self.classes > 0:
            data = np.array(display.canvas.axes.collections[0].get_offsets())
            size = data.shape[0]

            predicted = [0, 0, 0, 0]

            expected = display.canvas.axes.collections[0].get_gid()[0][0]

            for i in range(size):
                prediction = self.euclidean_classifier_point(data[i], self.lda)
                if prediction != expected:
                    predicted[prediction] = predicted[prediction] + 1
                    incorrect1 = incorrect1 + 1
                else:
                    predicted[prediction] = predicted[prediction] + 1

            text = "    Class 1:"

            for classes in range(self.classes):
                text = text + "{:12d}".format(predicted[classes])

            self.process.append(text)

        # computes the classification error for the second set
        #
        text = "    Class 2:"
        if self.classes > 1:
            data = np.array(display.canvas.axes.collections[1].get_offsets())
            size = data.shape[0]

            predicted = [0, 0, 0, 0]

            expected = display.canvas.axes.collections[1].get_gid()[0][0]

            for i in range(size):
                prediction = self.euclidean_classifier_point(data[i], self.lda)
                if prediction != expected:
                    predicted[prediction] = predicted[prediction] + 1
                    incorrect2 = incorrect2 + 1
                else:
                    predicted[prediction] = predicted[prediction] + 1

            for classes in range(self.classes):
                text = text + "{:12d}".format(predicted[classes])
            self.process.append(text)

        # computes the classification error for the third set
        #
        text = "    Class 3:"
        if self.classes > 2:
            data = np.array(display.canvas.axes.collections[2].get_offsets())
            size = data.shape[0]
            predicted = [0, 0, 0, 0]

            expected = display.canvas.axes.collections[2].get_gid()[0][0]

            for i in range(size):
                prediction = self.euclidean_classifier_point(data[i], self.lda)
                if prediction != expected:
                    predicted[prediction] = predicted[prediction] + 1
                    incorrect3 = incorrect3 + 1
                else:
                    predicted[prediction] = predicted[prediction] + 1

            for classes in range(self.classes):
                text = text + "{:12d}".format(predicted[classes])

            self.process.append(text)

        # computes the classification error for the fourth set
        #
        text = "    Class 4:"

        if self.classes > 3:
            data = np.array(display.canvas.axes.collections[3].get_offsets())
            size = data.shape[0]
            predicted = [0, 0, 0, 0]
            expected = display.canvas.axes.collections[3].get_gid()[0][0]

            for i in range(size):
                prediction = self.euclidean_classifier_point(data[i], self.lda)
                if prediction != expected:
                    predicted[prediction] = predicted[prediction] + 1
                    incorrect4 = incorrect4 + 1
                else:
                    predicted[prediction] = predicted[prediction] + 1

            for classes in range(self.classes):
                text = text + "{:12d}".format(predicted[classes])

            self.process.append(text)

        incorrect = incorrect1 + incorrect2 + incorrect3 + incorrect4

        error = (incorrect / samples) * 100

        text = "\n     Error Rate = {} / {} = {:.2f}%\n".format(incorrect, samples, error)

        self.process.append(text)

    def classify_eval(self):
        eval_classes = 0

        for collections in self.eval.canvas.axes.collections:
            if None is not collections.get_gid():
                eval_classes = eval_classes + 1

        eval_data = np.empty((0, 0))

        if eval_classes == 0:
            self.process.append("    There is no Eval Data to Classify.\n")
            return False

        elif eval_classes != self.classes:
            text = "    The number of classes in Eval Data ({}) does not match train data ({}).\n".format(eval_classes,
                                                                                                          self.classes)
            self.process.append(text)
            return False

        for collections in self.eval.canvas.axes.collections:
            # retrieve the x,y and gid of the collection
            #
            data = np.array(collections.get_offsets())

            eval_data = np.append(eval_data, data)

        eval_data = np.reshape(eval_data, (-1, 2))

        self.euclidean_classifier(eval_data, self.eval, eval_classes, self.lda)

        self.compute_errors(self.eval)

# class: AlgorithmLDA_cd
#
# This class implements the Class Dependent Linear Discrimination Anlaysis
# machine learning algorithm using the training and evaluation data currently
# stored in the app.
#
class AlgorithmLDACD():

    # method: AlgorithmLDA::constructor
    #
    # arguments:
    #  input_display: training window information 
    #  eval_display: evaluation window information
    #  process_box: process description window
    #
    def __init__(self,
                 input_display,
                 eval_display,
                 process_box):

        self.input = input_display
        self.eval = eval_display
        self.process = process_box
        self.class_num = np.empty((0, 0))
        self.X = np.empty((0, 0))
        self.classes = 0
        self.cov1 = 0
        self.cov2 = 0

        self.ldas = [LinearDiscriminantAnalysis(solver='eigen') for i in range(4)]
        self.support_region = np.empty((0, 0))

        self.LDA1_d = np.empty((2, 2)) * np.nan
        self.LDA2_d = np.empty((2, 2)) * np.nan
        self.LDA3_d = np.empty((2, 2)) * np.nan
        self.LDA4_d = np.empty((2, 2)) * np.nan

        self.trans_1 = None
        self.trans_2 = None
        self.trans_3 = None
        self.trans_4 = None
        self.trans_matrix = np.empty((0, 0))
        self.step_index = 0

        self.data = False

    def run(self):
        if self.step_index == 1:
            self.step1()

        elif self.step_index == 2:
            self.step2()
            self.process.append("\n   *** Algorithm Complete ***\n")

    def get_step_index(self):
        if self.step_index > 2:
            finished = True
        else:
            finished = False

        return finished

    def increment_step(self):
        self.step_index = self.step_index + 1

    def reset_step(self):
        self.step_index = 0

    def initialize(self):
        text = "Algorithm: Class Dependent Principal Component Analysis"
        self.process.append(text)
        text = "\n    *** Initializing the Algorithm ***\n"
        self.process.append(text)

        self.class_num = np.empty((0, 0))
        self.X = np.empty((0, 0))
        self.classes = 0

        self.ldas = [LinearDiscriminantAnalysis(solver='eigen') for i in range(4)]
        self.support_region = np.empty((0, 0))

        self.LDA1_d = np.empty((2, 2)) * np.nan
        self.LDA2_d = np.empty((2, 2)) * np.nan
        self.LDA3_d = np.empty((2, 2)) * np.nan
        self.LDA4_d = np.empty((2, 2)) * np.nan

        self.trans_1 = None
        self.trans_2 = None
        self.trans_3 = None
        self.trans_4 = None
        self.trans_matrix = np.empty((0, 0))

        self.step_index = 0
        self.data = False

    def step1(self):
        text = "    1. Computing the Statistics.\n" + "       Means:"
        self.process.append(text)

        self.print_means()
        self.input.canvas.draw_idle()
        self.print_matrices()

    def step2(self):
        text = "\n    2. Drawing the Decision Surface(s).\n "
        self.process.append(text)

        self.process.repaint()

        self.euclidean_classifier(self.X, self.input, self.classes, self.ldas)

        text = "    3. Computing Error Rates:\n "
        self.process.append(text)

        text = "    Train:"
        self.process.append(text)
        self.compute_errors(self.input)

        text = "    Eval:"
        self.process.append(text)
        self.classify_eval()

        self.input.reset_classes()

    def extract_data(self):
        self.classes = 0
        index = 0
        observed_classes = np.empty((0, 0))

        if len(self.input.canvas.axes.collections) == 0:
            return False
        else:
            self.data = True

        index_col = np.empty((0, 0), dtype="int64")
        collection_index = 0
        for collections in self.input.canvas.axes.collections:

            # retrieve the x,y and gid of the collection
            #
            try:
                current_gid = collections.get_gid()[0][0]
            except TypeError:
                current_gid = collections.get_gid()

            if current_gid is not None:
                data = np.array(collections.get_offsets())
                classes = np.array(collections.get_gid())

                self.class_num = np.append(self.class_num, classes)
                self.X = np.append(self.X, data)

                if current_gid not in observed_classes:
                    observed_classes = np.append(observed_classes, current_gid)
                    label = np.reshape(np.array(collections.get_gid()), (-1,))
                    self.ldas[index].fit_transform(data, label)
                    self.classes = self.classes + 1
                    index = index + 1
                    index_col = np.append(index_col, collection_index)

                else:
                    pca_index = 0
                    for collects in self.input.canvas.axes.collections:
                        if current_gid in collects.get_gid():
                            save_sets = collects.get_offsets()
                            current_set = collections.get_offsets()
                            final_set = np.vstack((save_sets, current_set))

                            save_gid = collects.get_gid()
                            cur_gid = collections.get_gid()
                            final_gid = np.hstack((save_gid, cur_gid))

                            collects.set_offsets(final_set)
                            collects.set_gid(final_gid)
                            break
                        else:
                            pca_index = pca_index + 1
            collection_index = collection_index + 1

        remove_collections = np.array(self.input.canvas.axes.collections)[index_col].tolist()
        self.input.canvas.axes.collections = remove_collections

        self.X = np.reshape(self.X, (-1, 2))

        self.class_num = np.reshape(self.class_num, (-1, 1))

        self.transformLDA(self.X, self.class_num, self.classes)

        return True

    def transformLDA(self, X, y, classes):

        index = np.where(y == 0)[0]
        if X[index].size > 0:
            eigVect = self.ldas[0].scalings_

            for j in range(2):
                for k in range(2):
                    self.LDA1_d[k][j] = eigVect[k][j]

            self.trans_matrix = np.append(self.trans_matrix, self.LDA1_d)

        index = np.where(y == 1)[0]
        if X[index].size > 0:
            eigVect = self.ldas[1].scalings_

            for j in range(2):
                for k in range(2):
                    self.LDA2_d[k][j] = eigVect[k][j]

            self.trans_matrix = np.append(self.trans_matrix, self.LDA2_d)

        index = np.where(y == 2)[0]
        if X[index].size > 0:
            eigVect = self.ldas[2].scalings_

            for j in range(2):
                for k in range(2):
                    self.LDA3_d[k][j] = eigVect[k][j]

            self.trans_matrix = np.append(self.trans_matrix, self.LDA3_d)

        index = np.where(y == 3)[0]
        if X[index].size > 0:
            eigVect = self.ldas[3].scalings_

            for j in range(2):
                for k in range(2):
                    self.LDA4_d[k][j] = eigVect[k][j]

            self.trans_matrix = np.append(self.trans_matrix, self.LDA4_d)


        self.trans_matrix = np.reshape(self.trans_matrix, (self.classes, 2, 2))

    def euclidean_classifier(self, X, display, classes, ldas):

        res = (display.canvas.axes.get_xlim()[1] - display.canvas.axes.get_ylim()[0]) / 100

        x_min, x_max = X[:, 0].min() - .75, X[:, 0].max() + .75
        y_min, y_max = X[:, 1].min() - .75, X[:, 1].max() + .75
        xx, yy = np.meshgrid(np.arange(x_min, x_max, res),
                             np.arange(y_min, y_max, res))

        Z = np.empty((0, 0))

        for pairs in np.c_[xx.ravel(), yy.ravel()]:
            distances = np.empty((0, 0))
            for data in range(classes):

                mean = np.reshape(ldas[data].means_,(-1,1))

                trans_mean = np.matmul(self.trans_matrix[data],mean)
                trans_pair = np.matmul(self.trans_matrix[data],pairs)

                eud = (distance.euclidean(trans_mean, trans_pair))
                distances = np.append(distances, eud)

            Z = np.append(Z, np.argmin(distances))

        Z = Z.reshape(xx.shape)

        display.canvas.axes.contourf(xx, yy, Z, alpha=0.4)
        display.canvas.draw_idle()

    def print_means(self):
        for means in range(self.classes):

            text = "       Class {}:{}".format(means + 1, np.reshape(self.ldas[means].means_[0], (2,)))

            self.input.canvas.axes.scatter(self.ldas[means].means_[0][0], self.ldas[means].means_[0][1], c='black', s=5)
            self.process.append(text)

    def print_matrices(self):

        text = "\n      Covariance matrix:"
        self.process.append(text)

        for cur_class in range(self.classes):
            cov = self.ldas[cur_class].covariance_

            a11 = str(round(cov[0][0], 2))
            a12 = str(round(cov[0][1], 2))
            a21 = str(round(cov[1][0], 2))
            a22 = str(round(cov[1][1], 2))

            text = ("      Class {}".format(cur_class + 1))
            self.process.append(text)

            text = ("         " + a11 + "    " + a12 + "\n" +
                    "        " + a21 + "     " + a22)

            self.process.append(text)

        text = ("\n      Transformation matrix:")
        self.process.append(text)

        for cur_class in range(self.classes):
            a11 = str(round(self.trans_matrix[cur_class][0][0], 2))
            a12 = str(round(self.trans_matrix[cur_class][0][1], 2))
            a21 = str(round(self.trans_matrix[cur_class][1][0], 2))
            a22 = str(round(self.trans_matrix[cur_class][1][1], 2))

            text = ("      Class {}".format(cur_class + 1))
            self.process.append(text)

            text = ("         " + a11 + "    " + a12 + "\n" +
                    "         " + a21 + "    " + a22)

            self.process.append(text)

    def euclidean_classifier_point(self, data, ldas):

        distances = np.empty((0, 0))
        for cur_class in range(self.classes):
            mean = np.reshape(ldas[cur_class].means_, (-1, 1))

            trans_mean = np.matmul(self.trans_matrix[cur_class], mean)
            trans_pair = np.matmul(self.trans_matrix[cur_class], data)

            eud = (distance.euclidean(trans_mean, trans_pair))
            distances = np.append(distances, eud)

        return np.argmin(distances)

    def compute_errors(self, display):
        # initialize variables
        #
        samples = 0

        incorrect1 = 0
        incorrect2 = 0
        incorrect3 = 0
        incorrect4 = 0

        for collection in range(self.classes):
            size = display.canvas.axes.collections[collection].get_offsets().shape[0]
            samples = samples + size

        text = "    Number of samples: {}".format(samples)
        self.process.append(text)

        if self.classes == 4:
            text = "    Ref/Hyp:    {:10s}{:10s}{:10s}{:10s}".format("Class 1", "Class 2", "Class 3", "Class 4")
        elif self.classes == 3:
            text = "    Ref/Hyp:    Class 1       Class 2        Class 3"

        elif self.classes == 2:
            text = "    Ref/Hyp:    {:10s}{:10s}".format("Class 1", "Class 2")

        elif self.classes == 1:
            text = "    Ref/Hyp:    {:10s}".format("Class 1")

        self.process.append(text)

        # computes the classification error for the first set
        #
        if self.classes > 0:
            data = np.array(display.canvas.axes.collections[0].get_offsets())
            size = data.shape[0]

            predicted = [0, 0, 0, 0]

            expected = display.canvas.axes.collections[0].get_gid()[0][0]

            for i in range(size):
                prediction = self.euclidean_classifier_point(data[i], self.ldas)
                if prediction != expected:
                    predicted[prediction] = predicted[prediction] + 1
                    incorrect1 = incorrect1 + 1
                else:
                    predicted[prediction] = predicted[prediction] + 1

            text = "    Class 1:"

            for classes in range(self.classes):
                text = text + "{:12d}".format(predicted[classes])

            self.process.append(text)

        # computes the classification error for the second set
        #
        text = "    Class 2:"
        if self.classes > 1:
            data = np.array(display.canvas.axes.collections[1].get_offsets())
            size = data.shape[0]

            predicted = [0, 0, 0, 0]

            expected = display.canvas.axes.collections[1].get_gid()[0][0]

            for i in range(size):
                prediction = self.euclidean_classifier_point(data[i], self.ldas)
                if prediction != expected:
                    predicted[prediction] = predicted[prediction] + 1
                    incorrect2 = incorrect2 + 1
                else:
                    predicted[prediction] = predicted[prediction] + 1

            for classes in range(self.classes):
                text = text + "{:12d}".format(predicted[classes])
            self.process.append(text)

        # computes the classification error for the third set
        #
        text = "    Class 3:"
        if self.classes > 2:
            data = np.array(display.canvas.axes.collections[2].get_offsets())
            size = data.shape[0]
            predicted = [0, 0, 0, 0]

            expected = display.canvas.axes.collections[2].get_gid()[0][0]

            for i in range(size):
                prediction = self.euclidean_classifier_point(data[i], self.ldas)
                if prediction != expected:
                    predicted[prediction] = predicted[prediction] + 1
                    incorrect3 = incorrect3 + 1
                else:
                    predicted[prediction] = predicted[prediction] + 1

            for classes in range(self.classes):
                text = text + "{:12d}".format(predicted[classes])

            self.process.append(text)

        # computes the classification error for the fourth set
        #
        text = "    Class 4:"

        if self.classes > 3:
            data = np.array(display.canvas.axes.collections[3].get_offsets())
            size = data.shape[0]
            predicted = [0, 0, 0, 0]
            expected = display.canvas.axes.collections[3].get_gid()[0][0]

            for i in range(size):
                prediction = self.euclidean_classifier_point(data[i], self.ldas)
                if prediction != expected:
                    predicted[prediction] = predicted[prediction] + 1
                    incorrect4 = incorrect4 + 1
                else:
                    predicted[prediction] = predicted[prediction] + 1

            for classes in range(self.classes):
                text = text + "{:12d}".format(predicted[classes])

            self.process.append(text)

        incorrect = incorrect1 + incorrect2 + incorrect3 + incorrect4

        error = (incorrect / samples) * 100

        text = "\n     Error Rate = {} / {} = {:.2f}%\n".format(incorrect, samples, error)

        self.process.append(text)

    def classify_eval(self):
        eval_classes = 0

        for collections in self.eval.canvas.axes.collections:
            if None is not collections.get_gid():
                eval_classes = eval_classes + 1

        eval_data = np.empty((0, 0))

        if eval_classes == 0:
            self.process.append("    There is no Eval Data to Classify.\n")
            return False

        elif eval_classes != self.classes:
            text = "    The number of classes in Eval Data ({}) does not match train data ({}).\n".format(eval_classes,
                                                                                                          self.classes)
            self.process.append(text)
            return False

        for collections in self.eval.canvas.axes.collections:
            # retrieve the x,y and gid of the collection
            #
            data = np.array(collections.get_offsets())

            eval_data = np.append(eval_data, data)

        eval_data = np.reshape(eval_data, (-1, 2))

        self.euclidean_classifier(eval_data, self.eval, eval_classes, self.ldas)

        self.compute_errors(self.eval)
