#!/usr/bin/env python3
#
# file: DataPoints.py
#
# This script defines preset patterns for data points including Two Gaussian,
# Four Gaussian, Overlapping Gaussian, Two Ellipses, Four Ellipses, Rotated
# Ellipses, Toroidal, and Yin and Yang
#-----------------------------------------------------------------------------
# import system modules
#
import numpy as np
import math

# input module for generating normal gaussian distribution
#
from scipy.stats import multivariate_normal

# class: DataPoints
#
# This class contains methods which each contain the data for one of several
# types of distributions that can be displayed in the Train and Eval windows
# using the Patterns menu.
#
class DataPoints:

    # method: DataPoints::constructor
    #
    # arguments:
    #  input_display: train window information
    #
    def __init__(self, input_display):

        # store reference to parent loop
        #
        self.input = input_display

        # yin-yang overlap parameter between [-1, 1]
        #
        self.overlap = 0.1 

    # method: DataPoints::set_two_gaussian
    #
    # arguments:
    #  max_size: number of points in distribution
    #  mu: mean matrix
    #  cov: covariance matrix
    #
    # This method creates and stores data of two normal gaussian distributions.
    #
    def set_two_gaussian(self, max_size=None, mu=None, cov=None):

        # makes sure that the size valid
        #
        if max_size is np.nan:

            # default to 10000
            #
            max_size = 10000

        # initialize local variables
        #
        x_min = self.input.canvas.axes.get_xlim()[0]
        x_max = self.input.canvas.axes.get_xlim()[1]
        y_min = self.input.canvas.axes.get_ylim()[0]
        y_max = self.input.canvas.axes.get_ylim()[1]
        class_zero = np.full((1, max_size), 0.0)
        class_one = np.full((1, max_size), 1.0)
        x_val = np.empty(max_size)
        y_val = np.empty(max_size)

        # generate the bi-normal gaussian random deviates
        #
        if True not in np.isnan(mu)[0] or True not in np.isnan(cov)[0]:
            gauss_plot = np.random.multivariate_normal(mu[0].transpose()[0],\
                                                       cov[0], max_size)

        else:
            mu_d = [x_min + (x_max - x_min) / 4.0,
                    y_max - (y_max - y_min) / 4.0]

            cov_d = [[((0.1 * (x_max - x_min)) / 4) * ((x_max - x_min) / 4),\
                      0.0],
                     [0.0,
                      ((0.1 * (y_max - y_min)) / 4) * ((y_max - y_min) / 4)]]

            gauss_plot = np.random.multivariate_normal(mu_d, cov_d, max_size)

        self.color_c = self.input.colors.pop()
        class_0 = self.input.canvas.axes.scatter(gauss_plot[:, 0], gauss_plot[:, 1],\
                                       s=1, gid=class_zero, color=self.color_c,)

        self.input.class_info["Class0"] = [class_0, gauss_plot[:, 0], gauss_plot[:, 1], True, self.color_c]

        if True not in np.isnan(mu)[1] or True not in np.isnan(cov)[1]:
            gauss_plot = np.random.multivariate_normal(mu[1].transpose()[0],\
                                                       cov[1], max_size)
        else:
            mu_d = [x_max - (x_max - x_min) / 4.0,
                    y_min + (y_max - y_min) / 4.0]

            cov_d = [[((0.1 * (x_max - x_min)) / 4) * ((x_max - x_min) / 4),\
                      0.0],
                     [0.0,\
                      ((0.1 * (y_max - y_min)) / 4) * ((y_max - y_min) / 4)]]

            gauss_plot = np.random.multivariate_normal(mu_d, cov_d, max_size)

        self.color_c = self.input.colors.pop()
        class_1 = self.input.canvas.axes.scatter(gauss_plot[:, 0], gauss_plot[:, 1],\
                                       s=1, gid=class_one, color =self.color_c)

        self.input.class_info["Class1"] = [class_1 ,gauss_plot[:, 0], gauss_plot[:, 1], True, self.color_c]

        self.input.canvas.draw_idle()

    # method: DataPoints::set_four_gaussian
    #
    # This method creates and stores data of four normal gaussian
    # distributions.
    #
    def set_four_gaussian(self, max_size=None, mu=None, cov=None):

        # makes sure that the size valid
        #
        if max_size is np.nan:

            # default to 10000
            #
            max_size = 10000

        # initialize local variables
        #
        x_min = self.input.canvas.axes.get_xlim()[0]
        x_max = self.input.canvas.axes.get_xlim()[1]
        y_min = self.input.canvas.axes.get_ylim()[0]
        y_max = self.input.canvas.axes.get_ylim()[1]

        # create four with dim (1, max_size) and fill
        #
        class_zero = np.full((1, max_size), 0)
        class_one = np.full((1, max_size), 1)
        class_two = np.full((1, max_size), 2)
        class_three = np.full((1, max_size), 3)


        # generate the bi-normal gaussian random deviates (first)
        #
        if True not in np.isnan(mu)[0] or True not in np.isnan(cov)[0]:
            gauss_plot = np.random.multivariate_normal\
                (mu[0].transpose()[0], cov[0], max_size)
        else:
            mu_d = [x_min + (x_max - x_min) / 4.0,
                    y_max - (y_max - y_min) / 4.0]

            cov_d = [[((0.1 * (x_max - x_min)) / 4) * ((x_max - x_min) / 4), \
                      0.0],
                     [0.0,\
                      ((0.1 * (y_max - y_min)) / 4) * ((y_max - y_min) / 4)]]

            gauss_plot = np.random.multivariate_normal(mu_d, cov_d, max_size)

        self.input.canvas.axes.scatter(gauss_plot[:, 0], gauss_plot[:, 1], \
                                       s=1, gid=class_zero)
        self.input.class_info["Class0"] = [True, gauss_plot[:, 0], gauss_plot[:, 1]]

        if True not in np.isnan(mu)[1] or True not in np.isnan(cov)[1]:
            gauss_plot = np.random.multivariate_normal(mu[1].transpose()[0], \
                                                       cov[1], max_size)
        else:
            mu_d = [x_max - (x_max - x_min) / 4.0,
                    y_min + (y_max - y_min) / 4.0]

            cov_d = [[((0.1 * (x_max - x_min)) / 4) * ((x_max - x_min) / 4), \
                      0.0],
                     [0.0, ((0.1 * (y_max - y_min)) / 4) * ((y_max - y_min) / \
                                                            4)]]

            gauss_plot = np.random.multivariate_normal(mu_d, cov_d, max_size)

        self.input.canvas.axes.scatter(gauss_plot[:, 0], gauss_plot[:, 1], \
                                       s=1, gid=class_one)
        self.input.class_info["Class1"] = [True, gauss_plot[:, 0], gauss_plot[:, 1]]

        if True not in np.isnan(mu)[2] or True not in np.isnan(cov)[2]:
            gauss_plot = np.random.multivariate_normal(mu[2].transpose()[0], \
                                                       cov[2], max_size)
        else:
            mu_d = [x_max - (x_max - x_min) / 4.0,
                    y_max - (y_max - y_min) / 4.0]

            cov_d = [[((0.1 * (x_max - x_min)) / 4) * ((x_max - x_min) / 4), \
                      0.0],
                     [0.0, ((0.1 * (y_max - y_min)) / 4) * ((y_max - y_min) / \
                                                            4)]]

            gauss_plot = np.random.multivariate_normal(mu_d, cov_d, max_size)

        self.input.canvas.axes.scatter(gauss_plot[:, 0], gauss_plot[:, 1], \
                                       s=1, gid=class_two)
        self.input.class_info["Class2"] = [True, gauss_plot[:, 0], gauss_plot[:, 1]]

        if True not in np.isnan(mu)[3] or True not in np.isnan(cov)[3]:
            gauss_plot = np.random.multivariate_normal(mu[3].transpose()[0], \
                                                       cov[3], max_size)
        else:
            mu_d = [x_min + (x_max - x_min) / 4.0,
                    y_min + (y_max - y_min) / 4.0]

            cov_d = [[((0.1 * (x_max - x_min)) / 4) * ((x_max - x_min) / 4), \
                      0.0],
                     [0.0, ((0.1 * (y_max - y_min)) / 4) * ((y_max - y_min) / \
                                                            4)]]

            gauss_plot = np.random.multivariate_normal(mu_d, cov_d, max_size)

        self.input.canvas.axes.scatter(gauss_plot[:, 0], gauss_plot[:, 1], \
                                       s=1, gid=class_three)
        self.input.class_info["Class3"] = [True, gauss_plot[:, 0], gauss_plot[:, 1]]

        # plot all the points
        #
        self.input.canvas.draw_idle()


    def set_over_gaussian(self, max_size=None, mu=None, cov=None):

        # makes sure that the size valid
        #
        if max_size is np.nan:
            
            # set default number of points to 10000
            #
            max_size = 10000
            
        # initialize local variables
        #
        x_min = self.input.canvas.axes.get_xlim()[0]
        x_max = self.input.canvas.axes.get_xlim()[1]
        y_min = self.input.canvas.axes.get_ylim()[0]
        y_max = self.input.canvas.axes.get_ylim()[1]
        class_zero = np.full((1, max_size), 0)
        class_one = np.full((1, max_size), 1)

        # generate the bi-normal gaussian random deviates
        #
        if True not in np.isnan(mu)[0] or True not in np.isnan(cov)[0]:
            gauss_plot = np.random.multivariate_normal(mu[0].transpose()[0], \
                                                       cov[0], max_size)
        else:
            mu_d = [(x_max - (x_max - x_min) / 1.75),
                    ((y_max - y_min) / 1.75 + y_min)]

            cov_d = [[((0.1 * (y_max - y_min)) / 4) * ((y_max - y_min) / 4), \
                      0.0],
                     [0.0, ((0.1 * (y_max - y_min)) / 4) * ((y_max - y_min) / \
                                                            4)]]

            gauss_plot = np.random.multivariate_normal(mu_d, cov_d, max_size)

        self.input.canvas.axes.scatter(gauss_plot[:, 0], gauss_plot[:, 1], \
                                       s=1, gid=class_zero)

        if True not in np.isnan(mu)[1] or True not in np.isnan(cov)[1]:
            gauss_plot = np.random.multivariate_normal(mu[1].transpose()[0], \
                                                       cov[1], max_size)
        else:

            mu_d = [((x_max - x_min) / 1.75 + x_min),
                    ((y_max - y_min) / 1.75 + y_min)]

            cov_d = [[((0.1 * (y_max - y_min)) / 4) * ((y_max - y_min) / 4), \
                      0.0],
                     [0.0, ((0.1 * (y_max - y_min)) / 4) * ((y_max - y_min) / \
                                                            4)]]

            gauss_plot = np.random.multivariate_normal(mu_d, cov_d, max_size)

        self.input.canvas.axes.scatter(gauss_plot[:, 0], gauss_plot[:, 1], \
                                       s=1, gid=class_one)

        self.input.canvas.draw_idle()

    def set_two_ellipses(self, max_size=None, mu=None, cov=None):
        # makes sure that the size valid
        #
        if max_size is np.nan:
            # default to 200
            #
            max_size = 10000
        # initialize local variables
        #
        x_min = self.input.canvas.axes.get_xlim()[0]
        x_max = self.input.canvas.axes.get_xlim()[1]
        y_min = self.input.canvas.axes.get_ylim()[0]
        y_max = self.input.canvas.axes.get_ylim()[1]
        class_zero = np.full((1, max_size), 0)
        class_one = np.full((1, max_size), 1)

        # generate the bi-normal gaussian random deviates
        #
        if True not in np.isnan(mu)[0] or True not in np.isnan(cov)[0]:
            gauss_plot = np.random.multivariate_normal(mu[0].transpose()[0], cov[0], max_size)
        else:
            mu_d = [x_min + (x_max - x_min) / 4.0,
                    y_max - (y_max - y_min) / 4.0]

            cov_d = [[((0.133 * (x_max - x_min)) / 4) * ((x_max - x_min) / 4),\
                      0.0],
                     [0.0, ((0.017 * (y_max - y_min)) / 4) * ((y_max - y_min)\
                                                              / 4)]]

            gauss_plot = np.random.multivariate_normal(mu_d, cov_d, max_size)

        self.input.canvas.axes.scatter(gauss_plot[:, 0], gauss_plot[:, 1], \
                                       s=1, gid=class_zero)

        if True not in np.isnan(mu)[1] or True not in np.isnan(cov)[1]:
            gauss_plot = np.random.multivariate_normal(mu[1].transpose()[0], \
                                                       cov[1], max_size)
        else:
            mu_d = [x_max - (x_max - x_min) / 4.0,
                    y_min + (y_max - y_min) / 4.0]

            cov_d = [[((0.133 * (x_max - x_min)) / 4) * ((x_max - x_min) / 4),\
                      0.0],
                     [0.0, ((0.017 * (y_max - y_min)) / 4) * ((y_max - y_min)\
                                                              / 4)]]

            gauss_plot = np.random.multivariate_normal(mu_d, cov_d, max_size)

        self.input.canvas.axes.scatter(gauss_plot[:, 0], gauss_plot[:, 1], \
                                       s=1, gid=class_one)

        self.input.canvas.draw_idle()

    def set_four_ellipses(self, max_size=None, mu=None, cov=None):

        # makes sure that the size valid
        #
        if max_size is np.nan:
            
            # set default number of points to 10000
            #
            max_size = 10000

        # initialize local variables
        #
        x_min = self.input.canvas.axes.get_xlim()[0]
        x_max = self.input.canvas.axes.get_xlim()[1]
        y_min = self.input.canvas.axes.get_ylim()[0]
        y_max = self.input.canvas.axes.get_ylim()[1]
        class_zero = np.full((1, max_size), 0)
        class_one = np.full((1, max_size), 1)
        class_two = np.full((1, max_size), 2)
        class_three = np.full((1, max_size), 3)

        # generate the bi-normal gaussian random deviates (first)
        #
        if True not in np.isnan(mu)[0] or True not in np.isnan(cov)[0]:
            gauss_plot = np.random.multivariate_normal(mu[0].transpose()[0], \
                                                       cov[0], max_size)
        else:
            mu_d = [x_min + (x_max - x_min) / 4.0,
                    y_max - (y_max - y_min) / 4.0]

            cov_d = [[((0.133 * (x_max - x_min)) / 4) * ((x_max - x_min) / 4),\
                      0.0],
                   [0.0, ((0.017 * (y_max - y_min)) / 4) * ((y_max - y_min)\
                                                            / 4)]]

            gauss_plot = np.random.multivariate_normal(mu_d, cov_d, max_size)

        self.input.canvas.axes.scatter(gauss_plot[:, 0], gauss_plot[:, 1], \
                                       s=1, gid=class_zero)

        if True not in np.isnan(mu)[1] or True not in np.isnan(cov)[1]:
            gauss_plot = np.random.multivariate_normal(mu[1].transpose()[0], \
                                                       cov[1], max_size)
        else:
            mu_d = [x_max - (x_max - x_min) / 4.0,
                    y_min + (y_max - y_min) / 4.0]

            cov_d = [[((0.133 * (x_max - x_min)) / 4) * ((x_max - x_min) / 4),\
                      0.0],
                   [0.0, ((0.017 * (y_max - y_min)) / 4) * ((y_max - y_min) / \
                                                            4)]]

            gauss_plot = np.random.multivariate_normal(mu_d, cov_d, max_size)

        self.input.canvas.axes.scatter(gauss_plot[:, 0], gauss_plot[:, 1], \
                                       s=1, gid=class_one)

        if True not in np.isnan(mu)[2] or True not in np.isnan(cov)[2]:
            gauss_plot = np.random.multivariate_normal(mu[2].transpose()[0], \
                                                       cov[2], max_size)
        else:
            mu_d = [x_max - (x_max - x_min) / 4.0,
                    y_max - (y_max - y_min) / 4.0]

            cov_d = [[((0.133 * (x_max - x_min)) / 4) * ((x_max - x_min) / 4),\
                      0.0],
                   [0.0, ((0.017 * (y_max - y_min)) / 4) * ((y_max - y_min) / \
                                                            4)]]

            gauss_plot = np.random.multivariate_normal(mu_d, cov_d, max_size)

        self.input.canvas.axes.scatter(gauss_plot[:, 0], gauss_plot[:, 1], \
                                       s=1, gid=class_two)

        if True not in np.isnan(mu)[3] or True not in np.isnan(cov)[3]:
            gauss_plot = np.random.multivariate_normal(mu[3].transpose()[0], \
                                                       cov[3], max_size)
        else:
            mu_d = [x_min + (x_max - x_min) / 4.0,
                    y_min + (y_max - y_min) / 4.0]

            cov_d = [[((0.133 * (x_max - x_min)) / 4) * ((x_max - x_min) / 4),\
                      0.0],
                   [0.0, ((0.017 * (y_max - y_min)) / 4) * \
                    ((y_max - y_min) / 4)]]

            gauss_plot = np.random.multivariate_normal(mu_d, cov_d, max_size)

        self.input.canvas.axes.scatter(gauss_plot[:, 0], gauss_plot[:, 1], \
                                       s=1, gid=class_three)

        self.input.canvas.draw_idle()

    def set_rotated_ellips(self,max_size=None,mu=None,cov=None):

        # makes sure that the size valid
        #
        if max_size is np.nan:
            
            # set default number of points to 10000
            #
            max_size = 10000

        # initialize local variables
        #
        x_min = self.input.canvas.axes.get_xlim()[0]
        x_max = self.input.canvas.axes.get_xlim()[1]
        y_min = self.input.canvas.axes.get_ylim()[0]
        y_max = self.input.canvas.axes.get_ylim()[1]
        class_zero = np.full((1, max_size), 0)
        class_one = np.full((1, max_size), 1)

        # generate the bi-normal gaussian random deviates
        #
        if True not in np.isnan(mu)[0] or True not in np.isnan(cov)[0]:
            gauss_plot = np.random.multivariate_normal(mu[0].transpose()[0], \
                                                       cov[0], max_size)
        else:
            mu_d = [x_min + (x_max - x_min) / 4.0,
                  y_max - (y_max - y_min) / 4.0]

            cov_d = [[((0.133 * (x_max - x_min)) / 4) * ((x_max - x_min) / 4),\
                      0.0],
                   [0.0, ((0.017 * (y_max - y_min)) / 4) * \
                    ((y_max - y_min) / 4)]]


            gauss_plot = np.random.multivariate_normal(mu_d, cov_d, max_size)

        self.input.canvas.axes.scatter(gauss_plot[:, 0], gauss_plot[:, 1], \
                                       s=1, gid=class_zero)

        if True not in np.isnan(mu)[1] or True not in np.isnan(cov)[1]:
            gauss_plot = np.random.multivariate_normal(mu[1].transpose()[0], \
                                                       cov[1], max_size)
        else:
            mu_d = [x_max - (x_max - x_min) / 4.0,
                  y_min + (y_max - y_min) / 4.0]

            cov_d = [[((0.017 * (x_max - x_min)) / 4) * ((x_max - x_min) / 4),\
                      0.0],
                   [0.0, ((0.133 * (x_max - x_min)) / 4) * \
                    ((x_max - x_min) / 4)]]

            gauss_plot = np.random.multivariate_normal(mu_d, cov_d, max_size)

        self.input.canvas.axes.scatter(gauss_plot[:, 0], gauss_plot[:, 1], \
                                       s=1, gid=class_one)

        self.input.canvas.draw_idle()

    def set_toroidal(self,params,mu,cov):

        # initialize local variables
        #
        x_min = self.input.canvas.axes.get_xlim()[0]
        x_max = self.input.canvas.axes.get_xlim()[1]
        y_min = self.input.canvas.axes.get_ylim()[0]
        y_max = self.input.canvas.axes.get_ylim()[1]

        # checks if user has specified parameters, otherwise defaults
        #
        if np.isnan(params[0]):
            
            # default total points to 10000
            #
            max_size = 10000
        else:
            max_size = int(params[0])

        if np.isnan(params[1]):
            
            # default number of points in ring to 10000
            #
            num_ring = 10000
        else:
            num_ring = int(params[1])

        if np.isnan(params[2]):

            # set default for the radius of the ring
            #
            ring_radius = 1.5 * (y_max - y_min) / 4
        else:
            ring_radius = params[2]

        if np.isnan(params[3]):

            # set default for the standard deviation of the ring
            #
            stddev_radius = 0.1 * (y_max - y_min) / 4
        else:
            stddev_radius = params[3]

        class_zero = np.full((1, num_ring), 0)
        class_one = np.full((1, max_size), 1)
        toroidal1 = np.empty((num_ring, 2))
        toroidal2 = np.empty((num_ring, 2))

        # create the outside ring
        #
        for i in range(1, num_ring + 1):
            
            # the angle is linearly distributed from 0 to 2Pi
            #
            angle = np.random.random() * 2 * math.pi
            radius = np.random.normal(ring_radius, stddev_radius)
            
            # set points
            #
            x_1 = radius * math.cos(angle)
            y_1 = radius * math.sin(angle)

            # plot the points
            #
            toroidal2[i - 1, 0] = x_1 + x_min + (x_max - x_min) / 2
            toroidal2[i - 1, 1] = y_1 + y_min + (y_max - y_min) / 2

        self.input.canvas.axes.scatter(toroidal2[:, 0], toroidal2[:, 1], \
                                       s=1, gid=class_zero)

        # generate the bi-normal gaussian random deviates
        #
        if True not in np.isnan(mu) or True not in np.isnan(cov):
            gauss_plot = np.random.multivariate_normal(mu.transpose()[0], \
                                                       cov, max_size)

        else:
            mu = [x_min + (x_max - x_min) / 2.0,
                  y_min + (y_max - y_min) / 2.0]

            cov = [[((0.033 * (x_max - x_min)) / 4) * ((x_max - x_min) / 4), \
                    0.0],
                   [0.0, ((0.033 * (x_max - x_min)) / 4) * \
                    ((x_max - x_min) / 4)]]

            gauss_plot = np.random.multivariate_normal(mu, cov, max_size)

        self.input.canvas.axes.scatter(gauss_plot[:, 0], gauss_plot[:, 1], \
                                       s=1, gid=class_one)

        self.input.canvas.draw_idle()

    # method: set_yin_yang
    #
    # arguments:
    #  parameters as retrieved from parameter pop-up
    #
    # returns:
    #  none
    #
    # This method creates the data for the Yin-Yang pattern using
    #  user-specified inputs or default values
    #
    def set_yin_yang(self, params):
        
        # initialize local variables
        #
        x_min = self.input.canvas.axes.get_xlim()[0]
        x_max = self.input.canvas.axes.get_xlim()[1]
        y_min = self.input.canvas.axes.get_ylim()[0]
        y_max = self.input.canvas.axes.get_ylim()[1]

        # the boundary, mean and standard deviation of plot
        #
        xmean = x_min + 0.5 * (x_max - x_min)
        ymean = y_min + 0.5 * (y_max - y_min)
        stddev_center = 1.5 * (x_max - x_min) / 2

        # creating empty lists to save coordinates of points
        #
        yin = []
        yang = []

        # calculate the radius of each class on the plot
        #
        radius1 = 1.5 * ((x_max - x_min) / 4)
        radius2 = 0.75 * ((x_max - x_min) / 4)

        # define number of samples in each class by checking for user-specified
        #  values and setting defaults if there are none
        #
        if np.isnan(params[0]):
            n_yin = 1000
        else:
            n_yin = int(params[0])
            
        if np.isnan(params[1]):
            n_yang = 1000
        else:
            n_yang = int(params[1])

        class_zero = np.full((1, 1000), 0)
        class_one = np.full((1, 1000), 1)

        # producing some random numbers based on Normal distribution and then
        # calculating the points distance to each class, choosing the closest
        # set.
        # the look will exit when both classes has been built up.
        #
        n_yin_counter = 0
        n_yang_counter = 0
        
        while ((n_yin_counter < n_yin) |
               (n_yang_counter < n_yang)):

            # generate points with Normal distribution
            #
            xpt = np.random.normal(xmean, stddev_center, 1)[0]
            ypt = np.random.normal(ymean, stddev_center, 1)[0]

            # calculate radius for each generated point
            #
            distance1 = np.sqrt(xpt ** 2 + ypt ** 2)
            distance2 = np.sqrt(xpt ** 2 + (ypt + radius2) ** 2)
            distance3 = np.sqrt(xpt ** 2 + (ypt - radius2) ** 2)

            # decide which class each point belongs to.
            # when the point added to each class, its counter increases by 1.
            # if the counter reach to the requested number of samples in
            # each class, then no sample will appended to that class and the
            # produced point will be thrown away.
            #
            if distance1 <= radius1:
                if (xpt >= -radius1) & (xpt <= 0):
                    if (((distance1 <= radius1) |
                         (distance2 <= radius2)) &
                            (distance3 > radius2)):
                        if n_yin_counter < n_yin:
                            yin.append([xpt, ypt])
                            n_yin_counter += 1
                    elif n_yang_counter < n_yang:
                        yang.append([xpt, ypt])
                        n_yang_counter += 1
                if (xpt > 0.0) & (xpt <= radius1):
                    if (((distance1 <= radius1) |
                         (distance3 <= radius2)) &
                            (distance2 > radius2)):
                        if n_yang_counter < n_yang:
                            yang.append([xpt, ypt])
                            n_yang_counter += 1
                    elif n_yin_counter < n_yin:
                        yin.append([xpt, ypt])
                        n_yin_counter += 1

        # translate each sample in Yin and Yang from the origin to
        # the center of the plot.
        # for implementing overlap, the overlap parameter multiply to one of
        # the plot center points. So the overlap parameter interferes in
        # translation process.
        #
        yang = np.array(yang) + \
               np.array([xmean, ymean])

        yin = np.array(yin) + \
              np.array([xmean, ymean]) * \
              (1 + self.overlap)

        self.input.canvas.axes.scatter(yang[:, 0], yang[:, 1],
                                       alpha=0.5, s=3, gid=class_zero)
        self.input.canvas.axes.scatter(yin[:, 0], yin[:, 1],
                                       alpha=0.5, s=3, gid=class_one)

        self.input.canvas.draw_idle()
